"""

@author: Johannes Kolberg
"""

import numpy as np
import pandas as pd
from scipy import signal


class TruncatedGaussian:
    """Right-truncated Gaussian signal processing filter to smooth time series values,
    """

    def __init__(self, lookback_window=None, sigma=None, threshold=None):
        """
        :param lookback_window: Number of preceding timestamps to include in the window.
        :param sigma: Standard deviation of the complete Gaussian distribution.
        :param (optional) threshold: Values above are set to 1, those below to 0.
        """
        self.threshold = threshold
        if lookback_window or sigma:
            if lookback_window:
                full_distribution_datapoints = lookback_window * 2 + 1
                sigma = sigma or 0.28 * lookback_window
            else:
                full_n = int((2 / 0.28) * sigma + 1)
                full_distribution_datapoints = full_n if full_n % 2 == 1 else full_n + 1
                lookback_window = int((full_distribution_datapoints - 1) / 2)
            self.kernel = np.array(signal.gaussian(full_distribution_datapoints, sigma))
            self.kernel = self.kernel[0 : lookback_window + 1]
            self.kernel = self.kernel / sum(self.kernel)
        else:
            self.kernel = None

    def fit(self, model_predictions, allowed_flags=0):
        """Fitting kernel to predictions to achieve defined maximum number of flags.
        If the window size or sigma exist, the kernel also does, so it is simply returned.
        Note that this ignores the maximum number of allowed flags.
        If the kernel does not exist, fit it to achieve that condition.

        :param model_predictions: Pandas df where first col predictions, index is timestamps.
        :param allowed_flags: Maximum number of value 1.0 the kernel is allowed to output
        :return: The fitted kernel.
        """

        # Kernel exists if window or sigma was defined, so there's no need to fit
        if self.kernel:
            return self
        # TODO: Raise warning if using default of 0.99
        if not self.threshold:
            self.threshold = 0.99
        # TODO: Set some stopping condition
        lookback = 0
        while True:
            # TODO: Better/more intelligent way of searching for window sizes
            lookback += 10
            if lookback % 50 == 0:
                print("Trying window of {} datapoints".format(lookback))
            kernel = np.array(signal.gaussian(1 + lookback * 2, 0.28 * lookback))
            kernel = kernel[0 : lookback + 1]
            self.kernel = kernel / sum(kernel)
            kernel_flags = self.transform(model_predictions)
            flagged = kernel_flags.iloc[:, 0].sum()
            if flagged <= allowed_flags:
                print("Found window of {} with {} anomalies flagged".format(lookback, flagged))
                break
        return self

    def transform(self, model_predictions):
        """Apply a defined or fitted kernel to a vector of values.

        :param model_predictions: Pandas df where first col predictions, index is timestamps.
        :return: A corresponding Pandas df with the resulting values.
        """
        # Assumes that the input model_predictions has timestamps as index
        kernel_scores = pd.Series(index=model_predictions.index)
        for i in range(len(model_predictions)):
            preds = model_predictions.iloc[:, 0].values[max(0, i + 1 - len(self.kernel)) : i + 1]
            preds = np.pad(preds, (len(self.kernel) - len(preds), 0), "constant")
            score = np.dot(self.kernel, preds)
            kernel_scores.at[model_predictions.index[i]] = score
        # Apply the threshold if it is set, to get binary [0, 1] anomaly flags
        if self.threshold:
            kernel_scores = kernel_scores.apply(lambda y: 1 if y >= self.threshold else 0)
        # Otherwise return the continuous [0, 1] anomaly scores
        return kernel_scores.to_frame(name="kernelScore")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

@author: quetran
"""
import unittest

import numpy as np
import pandas as pd
from pandas.util.testing import assert_frame_equal

from cognite_ml.timeseries.pattern_search import auxiliary
from nose_parameterized import parameterized


class AuxTestCase(unittest.TestCase):
    fill_nan_testcases = [
        # Test ffill
        (pd.DataFrame([[None, 2, 3, 4], [2, 2, 3, 4]]), 0),
        # Test bfill
        (pd.DataFrame([[2, 2, 3, 4], [None, 2, 3, 4]]), 0),
        # Test default values
        (pd.DataFrame([[None, 2, 3, 4], [None, 2, 3, 4]]), 2.0),
    ]

    @parameterized.expand(fill_nan_testcases)
    def test_fill_nan(self, df, default_value):
        expected_output = pd.DataFrame([[2.0, 2, 3, 4], [2.0, 2, 3, 4]])
        result = auxiliary.fill_nan(df, default_value)
        assert_frame_equal(expected_output, result)

    def test_convert_time_delta_to_num_points(self):
        granularity = 10 * 60
        time_delta = "1h"
        assert auxiliary.convert_time_delta_to_num_points(granularity, time_delta) == 7

    def test_convert_index(self):
        stride = 3
        ts_indices = np.array([0, 3, 6, 9, 12])
        word_indices = np.array([0, 1, 2, 3, 4])
        assert np.array_equal(auxiliary.convert_index(stride=stride, word_indices=word_indices), ts_indices)
        assert np.array_equal(auxiliary.convert_index(stride=stride, ts_indices=ts_indices), word_indices)
        self.assertRaises(ValueError, auxiliary.convert_index, stride)

    def mapping_timeseries(self, common_aggregates_input, common_aggregates_query):
        input_timeseries = {
            "timeSeries": [
                {"name": "ts1", "missingDataStrategy": "ffill"},
                {"name": "ts2", "aggregates": ["step"], "missingDataStrategy": "ffill"},
            ],
            "aggregates": common_aggregates_input,
        }

        query_timeseries = {
            "timeSeries": [
                {"name": "ts1", "aggregates": ["cv", "dv"], "missingDataStrategy": "ffill"},
                {"name": "ts2", "aggregates": ["step"], "missingDataStrategy": "ffill"},
            ],
            "aggregates": common_aggregates_query,
        }

        return input_timeseries, query_timeseries

    def test_whiten_univariate(self):
        window_signal = [1, 1, 1, 1, 1]
        self.assertTrue(np.array_equal(auxiliary.whiten_univariate(window_signal), np.array([0, 0, 0, 0, 0])))

    # TODO: Should we add test for find_variance_thresholds?

    calculate_overlap_testcases = [
        # Test not overlap
        (5, 10, 1, 4, 0.0),
        # Test not overlap
        (5, 10, 12, 15, 0.0),
        # Test overlap left side
        (5, 10, 1, 5, 0.2),
        # Test overlap right side
        (5, 10, 8, 13, 0.5),
        # Test new period inside old period
        (5, 10, 6, 9, 1),
        # Test old period inside new period
        (5, 10, 3, 10, 0.75),
    ]

    @parameterized.expand(calculate_overlap_testcases)
    def test_calculate_overlap(self, start_old, end_old, start_new, end_new, expected_result):
        self.assertEqual(auxiliary.calculate_overlap(start_old, end_old, start_new, end_new), expected_result)

    extact_matches_testcases = [
        # Test without overlapping
        (None, [{"from": 3, "to": 7, "score": 1}, {"from": 6, "to": 8, "score": 2}, {"from": 2, "to": 5, "score": 3}]),
        # Test with overlapping
        (0.7, [{"from": 3, "to": 7, "score": 1}, {"from": 6, "to": 8, "score": 2}]),
    ]

    @parameterized.expand(extact_matches_testcases)
    def test_extact_matches(self, overlap_ratio, expected_matches):
        timestamp_list = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        sorted_distance = [
            {"distance": 1, "stride": 3, "len": 5, "i": 1},
            {"distance": 2, "stride": 1, "len": 3, "i": 6},
            {"distance": 3, "stride": 2, "len": 4, "i": 1},
        ]
        matches = auxiliary.extract_matches(sorted_distance, timestamp_list, overlap_ratio, 3)
        self.assertCountEqual(expected_matches, matches)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

@author: quetran
"""
import unittest

import numpy as np
import pandas as pd
from cognite_ml.timeseries.pattern_search.algorithms.DTW.searcher import DTW


class DTWSearcherTestCase(unittest.TestCase):
    def test_find_similar_patterns(self):
        search_len = 3
        expected_seq = pd.DataFrame([[2, 2], [2, 2], [4, 3], [4, 3], [4, 4], [4, 4]], dtype=np.float64)
        pattern = pd.DataFrame([[2, 2], [4, 3], [4, 4]], dtype=np.float64)
        timeseries = pd.concat([pattern, pattern * 100])

        searcher = DTW(min_range=None, max_range=None, limit=5)

        res = searcher.find_similar_patterns(search_len, expected_seq.values, timeseries.values)
        self.assertEqual(searcher.limit, 5)
        self.assertEqual(searcher.sakoeChibaWidth, 0.1)

        top = sorted(res, key=lambda dict: dict["distance"])[0:2]
        for item in top:
            item["distance"] = round(item["distance"], 3)
        expected_res = [{"distance": 0.0, "stride": 1, "len": 3, "i": 0, "window": slice(0, 3, None)}]
        self.assertCountEqual(top, expected_res)

    def test_search(self):
        timestamp = pd.DataFrame((np.linspace(1000, 6000, 6)).astype(int), columns=["timestamp"])
        expected_seq = pd.DataFrame(
            [[2, 2], [2, 2], [4, 3], [4, 3], [4, 4], [4, 4]], columns=["a", "b"], dtype=np.float64
        ).reset_index(drop=True)
        expected_seq = expected_seq.join(timestamp)
        pattern = pd.DataFrame([[2, 2], [4, 3], [4, 4]], columns=["a", "b"], dtype=np.float64)
        timeseries = pd.concat([pattern, pattern * 100]).reset_index(drop=True)
        timeseries = timeseries.join(timestamp)

        searcher = DTW(min_range="2s", max_range=None, limit=2)
        searcher.filterVariance = False

        matches = searcher.search(timeseries, expected_seq, None)
        for m in matches:
            m["score"] = round(m["score"], 3)

        expected_res = [
            {"from": 1000, "to": 3000, "score": 0.0, "percentage": 100.0},
            {"from": 4000, "to": 6000, "score": 0.0, "percentage": 100.0},
        ]
        self.assertCountEqual(expected_res, matches)

    def test_is_overlapping(self):
        searcher = DTW(min_range=None, max_range=None, limit=5)
        searcher.overlap_th = 0.7
        matches = [{"from": 0, "to": 5, "score": 0}, {"from": 8, "to": 10, "score": 0}]
        self.assertFalse(searcher._is_overlapping(matches, 6, 7))
        self.assertTrue(searcher._is_overlapping(matches, 8, 9))
        self.assertFalse(searcher._is_overlapping(matches, 5, 6))  # 50% overlap should be accepted as not overlapping

    def test_lb_to_dtw_distance(self):
        timestamp = pd.DataFrame((np.linspace(1000, 6000, 6)).astype(int), columns=["timestamp"])
        expected_seq = pd.DataFrame(
            [[2, 2], [2, 2], [4, 3], [4, 3], [4, 4], [4, 4]], columns=["a", "b"], dtype=np.float64
        ).reset_index(drop=True)
        expected_seq = expected_seq.join(timestamp)
        pattern = pd.DataFrame([[2, 2], [4, 3], [4, 4]], columns=["a", "b"], dtype=np.float64)
        timeseries = pd.concat([pattern, pattern * 100]).reset_index(drop=True)
        timeseries = timeseries.join(timestamp)

        searcher = DTW(min_range=None, max_range=None, limit=2)
        searcher.input_ts = timeseries
        searcher.query_seq = expected_seq
        input_ts_cols = ["a", "b"]
        query_seqs_cols = ["a", "b"]
        timestamp = timestamp.reset_index()["timestamp"]

        matches = [{"from": 1000, "to": 6000, "score": 0}]
        dict = {"distance": 3, "stride": 1, "len": 4, "i": 0, "window": slice(0, 4, None)}

        op = searcher._lb_to_dtw_distance(
            dict,
            matches,
            timestamp,
            expected_seq.as_matrix(columns=query_seqs_cols),
            timeseries.as_matrix(columns=input_ts_cols),
        )

        self.assertEqual(op, {"from": 1000, "to": 4000, "score": np.inf})  # Distance should be inf when overlap
        matches = [{"from": 5000, "to": 6000, "score": 0}]

        op = searcher._lb_to_dtw_distance(
            dict,
            matches,
            timestamp,
            expected_seq.as_matrix(columns=query_seqs_cols),
            timeseries.as_matrix(columns=input_ts_cols),
        )

        self.assertCountEqual(op, {"from": 1000, "to": 4000, "score": None})
        self.assertGreaterEqual(op["score"], 0)
        self.assertNotEquals(op["score"], np.inf)  # Distance should not be inf when no overlap



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

@author: agleim
"""

import numpy as np
from sklearn.utils.validation import check_array
from sklearn.base import BaseEstimator, TransformerMixin


__all__ = ["SlidingWindows"]


class SlidingWindows(TransformerMixin):
    def __init__(self, window_size=2, step_size=1):
        self.window_size = window_size
        self.step_size = step_size

    def fit(self, *_):
        return self

    def transform(self, X, *_):

        X = check_array(X)

        (num_samples, num_dims) = X.shape
        window_size = self.window_size
        step_size = self.step_size

        assert num_samples >= window_size, "The window size can't exceed number of samples."

        windows = np.zeros((int(np.ceil((num_samples - window_size + 1) / step_size)), num_dims * window_size))

        for i in range(0, num_samples, step_size):
            if i + window_size > num_samples:
                break

            w = X[i : i + window_size, :].ravel()
            windows[int(np.ceil(i / step_size)), :] = w

        return windows



import pytest
from cognite_ml.metrics import RangeBasedPrecisionRecall


class TestClassRangeBasedPrecisionRecall(object):

    def test_empty_input(self):
        anomaly_labels = []
        predicted_labels = []

        pm = RangeBasedPrecisionRecall(
            anomaly_labels=anomaly_labels,
            predicted_labels=predicted_labels,
            alpha=0.5,
            beta=0.5,
            bias_type="front",
            granularity="10m",
        )
        with pytest.raises(ZeroDivisionError):
            pm.get_recall()




import numpy as np
import pandas as pd


class ADFunctionalBase:
    def __init__(
        self,
        base_pipeline,
        train_window_size=None,
        growing_train_window=False,
        predict_window_size=None,
        num_lags=0,
        outlier_filtering="3sigma",
    ):

        self.base_pipeline = base_pipeline
        self.train_size = train_window_size
        self.predict_size = predict_window_size
        self.growing_train_window = growing_train_window
        self.filtering = outlier_filtering
        self.num_lags = 0

    def _fit_window(self, data):
        y = data[self.target_var]
        X = data.drop([self.target_var, "timestamp"], axis=1)

        self.base_pipeline.fit(X, y)

    def _predict_window(self, data):
        T = data.timestamp.values
        y_true = data[self.target_var].values
        X = data.drop([self.target_var, "timestamp"], axis=1)

        y_pred = self.base_pipeline.predict(X)

        self.residual.loc[T, "Predicted"] = y_pred
        self.residual.loc[T, "Residual"] = y_true - y_pred

        return 0

    def _calc_score(self):
        L2 = np.sum((1.0 / len(self.residual)) * np.abs(self.residual["Residual"].values) ** 2) ** 0.5

        self.score = {}
        self.score["L2"] = L2

        return 0

    def fit(self, data, target_var):
        self.target_var = target_var
        self.residual = pd.DataFrame(columns=["Predicted", "Residual"], index=data.timestamp)

        T = data.timestamp.values
        Tstart = T[0]
        Tstop = T[-1]
        DT_training = int(pd.Timedelta(self.train_size).total_seconds() * 1000)
        DT_predict = int(pd.Timedelta(self.predict_size).total_seconds() * 1000)

        Tcurrent = Tstart + DT_training
        while Tcurrent < Tstop:
            if self.growing_train_window:
                train_start = Tstart
            else:
                train_start = Tcurrent - DT_training
            train_end = Tcurrent
            predict_start = Tcurrent
            predict_end = Tcurrent + DT_predict

            print("Training period:", pd.to_datetime(train_start, unit="ms"), pd.to_datetime(predict_end, unit="ms"))

            training_window = (data.timestamp > train_start) & (data.timestamp < train_end)
            predict_window = (data.timestamp > predict_start) & (data.timestamp < predict_end)

            data_train = data[training_window]
            data_predict = data[predict_window]

            self._fit_window(data_train)
            self._predict_window(data_predict)

            Tcurrent = predict_end

        self.valid_to = Tstop + DT_predict

        self.residual["timestamp"] = self.residual.index
        self.residual = self.residual.reset_index(drop=True)
        self.residual = self.residual.fillna(value=0.0)
        self._calc_score()

        return self.score

    def predict(self, data):
        print("not implemented")
        return -1



# distutils: language=c++

import numpy as np
cimport numpy as np
from libcpp.vector cimport vector




def __exponential(phi, z, K):
    cdef int num_points = len(z)

    for i in range(num_points):
        phi[i] = np.exp(-np.abs(z[i])**2/K)

    return 1



def __classic(phi, z, K):
    cdef int num_points = len(z)
    cdef double tmp

    for i in range(num_points):
        tmp = 1.0 + np.abs(z[i]/K)**2
        phi[i] = 1.0/tmp

    return 1


def __tukey(phi, z, K):
    cdef int num_points = len(z)

    for i in range(num_points):
        if(np.abs(z[i]) <= K):
            phi[i] = 0.5*(1.0 - np.abs(z[i]/K)**2)**2
        else:
            phi[i] = 0.0

        
    return 1


def __guo(phi, z, K):
    cdef int num_points = len(z)
    cdef double tmp
    cdef double alpha
    cdef double beta

    for i in range(num_points):
        beta = 1.0 + np.abs(z[i]/K)**2
        alpha = 2 - 2/beta
        phi[i] = 1.0 + np.abs(z[i]/K)**alpha

    return 1


def __weikert(phi, z, K):
    cdef int num_points = len(z)
    cdef double tmp
    cdef double alpha
    cdef double beta

    for i in range(num_points):
        if(np.abs(z[i]) > 0.0000000001):
            phi[i] = 1.0 - np.exp(-3.31488*K**8/np.abs(z[i])**8)
        else:
            phi[i] = 0.0

    return 1



def __solve_tri_linear(a, b, c, d, x):

    cdef int num_points = len(a)
    cdef double w

    #Forward sweep
    for i in range(1, num_points):
        w = a[i]/b[i-1]
        b[i] = b[i] - w*c[i-1]
        d[i] = d[i] - w*d[i-1] 

    #Backward sweep
    x[num_points-1] = d[num_points-1]/b[num_points-1]
    for i in range(num_points-2, -1, -1):
        x[i] = (1.0/b[i])*(d[i] - c[i]*x[i+1])

    return 1

def __perona_malik_implicit(X, t=None, smoothing_factor=1, gradient_threshold=1, method='classic'):
    cdef int num_points = len(X)

    cdef np.ndarray[np.double_t, ndim=1] Y      # The returned function
    cdef np.ndarray[np.double_t, ndim=1] phi    # Gradient stop function
    cdef np.ndarray[np.double_t, ndim=1] dx     # Gradient at i+0.5
    cdef np.ndarray[np.double_t, ndim=1] dt     # Time diff i+0.5

    cdef np.ndarray[np.double_t, ndim=1] Adia     # Matrix diagonal
    cdef np.ndarray[np.double_t, ndim=1] Adiap    # Matrix sup-diagonal
    cdef np.ndarray[np.double_t, ndim=1] Adiam    # Matrix sub-diagonal
    cdef np.ndarray[np.double_t, ndim=1] b        # Linear system RHS
    
    # Tmp storage we will need later
    cdef double DT
    cdef double gammaP
    cdef double gammaM
    
    # Let us allocate space for the vectors
    Y     = np.zeros(num_points)
    phi   = np.zeros(num_points)
    dx    = np.zeros(num_points-1)
    dt    = np.ones(num_points-1)
    Adia  = np.zeros(num_points)
    Adiap = np.zeros(num_points)
    Adiam = np.zeros(num_points)
    b     = np.zeros(num_points)

    # Now, let us calculate gradients
    if(type(t) != type(None)):
        for i in range(0, num_points-1):
            dt[i] = t[i+1]-t[i]

    for i in range(0, num_points-1):
        dx[i] = (X[i+1]-X[i])/dt[i]

    # Next, we calculate the stop gradient
    if(method=='classic'):
        __classic(phi, dx, smoothing_factor)
    elif(method=='exponential'):
        __exponential(phi, dx, smoothing_factor)
    elif(method=='tukey'):
        __tukey(phi, dx, smoothing_factor)
    elif(method=='guo'):
        __guo(phi, dx, smoothing_factor)
    elif(method=='weikert'):
        __weikert(phi, dx, smoothing_factor)

    DT = smoothing_factor
    
    #--- Now build the linear system ---
    
    # We begin with the internal points 
    cdef double theta = 1.0
    for i in range(1, num_points-1):
        gammaP = 2*DT/(dt[i]*(dt[i-1]+dt[i]))
        gammaM = 2*DT/(dt[i-1]*(dt[i-1]+dt[i]))
        Adia[i]  = (1.0 + gammaP*theta*phi[i] + gammaM*theta*phi[i-1])
        Adiap[i] = -theta*gammaP*phi[i]
        Adiam[i] = -theta*gammaM*phi[i-1]
        b[i] = X[i] + (1-theta)*(gammaP*phi[i]*(X[i+1]-X[i]) - gammaM*phi[i-1]*(X[i]-X[i-1]))

    # Then the boundary
    Adia[0] = 1.0
    Adiap[0]= -1.0
    b[0]    = 0.0
    Adia[num_points-1] = 1.0
    Adiam[num_points-1] = -1.0
    b[num_points-1] = 0.0
     
    # Next, we solve the linear system
    __solve_tri_linear(Adiam, Adia, Adiap,  b, Y)

    return Y



def __perona_malik_explicit(X, t=None, smoothing_factor=1, gradient_threshold=1, method='classic'):
    cdef int num_points = len(X)

    cdef np.ndarray[np.double_t, ndim=1] Y      # The returned function
    cdef np.ndarray[np.double_t, ndim=1] phi    # Gradient stop function
    cdef np.ndarray[np.double_t, ndim=1] dx     # Gradient at i+0.5
    cdef np.ndarray[np.double_t, ndim=1] dt     # Time diff i+0.5

    cdef np.ndarray[np.double_t, ndim=1] Adia     # Matrix diagonal
    cdef np.ndarray[np.double_t, ndim=1] Adiap    # Matrix sup-diagonal
    cdef np.ndarray[np.double_t, ndim=1] Adiam    # Matrix sub-diagonal
    cdef np.ndarray[np.double_t, ndim=1] b        # Linear system RHS
    
    # Tmp storage we will need later
    cdef double DT
    cdef double gammaP
    cdef double gammaM
    
    # Let us allocate space for the vectors
    Y     = np.zeros(num_points)
    phi   = np.zeros(num_points)
    dx    = np.zeros(num_points-1)
    dt    = np.ones(num_points-1)
    Adia  = np.zeros(num_points)
    Adiap = np.zeros(num_points)
    Adiam = np.zeros(num_points)
    b     = np.zeros(num_points)

    # Now, let us calculate gradients
    if(type(t) != type(None)):
        for i in range(0, num_points-1):
            dt[i] = t[i+1]-t[i]
    
    for i in range(0, num_points):
        Y[i] = X[i]

    cdef int k=0
    DT = smoothing_factor
    while(k<DT):

        for i in range(0, num_points-1):
            dx[i] = (Y[i+1]-Y[i])/dt[i]

        # Next, we calculate the stop gradient
        if(method=='classic'):
            __classic(phi, dx, smoothing_factor)
        elif(method=='exponential'):
            __exponential(phi, dx, smoothing_factor)
        elif(method=='tukey'):
            __tukey(phi, dx, smoothing_factor)
        elif(method=='guo'):
            __guo(phi, dx, smoothing_factor)
        elif(method=='weikert'):
            __weikert(phi, dx, smoothing_factor)

        
    
        for i in range(1, num_points-1):
            gammaP = .5/(dt[i]*(dt[i-1]+dt[i]))
            gammaM = .5/(dt[i-1]*(dt[i-1]+dt[i]))
            Y[i] = Y[i] + (gammaP*phi[i]*(Y[i+1]-Y[i]) - gammaM*phi[i-1]*(Y[i]-Y[i-1]))

        Y[0] = Y[1]
        Y[num_points-1] = Y[num_points-2]
        k = k +1

    return Y


def __perona_malik_midpoint(X, t=None, smoothing_factor=1, gradient_threshold=1, method='classic'):
    cdef int num_points = len(X)

    cdef np.ndarray[np.double_t, ndim=1] Y      # The returned function
    cdef np.ndarray[np.double_t, ndim=1] phi    # Gradient stop function
    cdef np.ndarray[np.double_t, ndim=1] phi_next    # Gradient stop function
    cdef np.ndarray[np.double_t, ndim=1] dx     # Gradient at i+0.5
    cdef np.ndarray[np.double_t, ndim=1] dx_next     # Gradient at i+0.5
    cdef np.ndarray[np.double_t, ndim=1] dt     # Time diff i+0.5

    cdef np.ndarray[np.double_t, ndim=1] Adia     # Matrix diagonal
    cdef np.ndarray[np.double_t, ndim=1] Adiap    # Matrix sup-diagonal
    cdef np.ndarray[np.double_t, ndim=1] Adiam    # Matrix sub-diagonal
    cdef np.ndarray[np.double_t, ndim=1] b        # Linear system RHS
    
    # Tmp storage we will need later
    cdef double DT
    cdef double gammaP
    cdef double gammaM
    
    # Let us allocate space for the vectors
    Y     = np.zeros(num_points)
    phi   = np.zeros(num_points)
    phi_next   = np.zeros(num_points)
    dx    = np.zeros(num_points-1)
    dx_next    = np.zeros(num_points-1)
    dt    = np.ones(num_points-1)
    Adia  = np.zeros(num_points)
    Adiap = np.zeros(num_points)
    Adiam = np.zeros(num_points)
    b     = np.zeros(num_points)

    # Now, let us calculate gradients
    if(type(t) != type(None)):
        for i in range(0, num_points-1):
            dt[i] = t[i+1]-t[i]

    Xnext = __perona_malik_implicit(X, t, smoothing_factor, gradient_threshold, method)
    
    for i in range(0, num_points-1):
        dx_next[i] = (Xnext[i+1]-Xnext[i])/dt[i]

    # Next, we calculate the stop gradient
    if(method=='classic'):
        __classic(phi, dx, smoothing_factor)
        __classic(phi_next, dx_next, smoothing_factor)
    elif(method=='exponential'):
        __exponential(phi, dx, smoothing_factor)
        __exponential(phi_next, dx_next, smoothing_factor)
    elif(method=='tukey'):
        __tukey(phi, dx, smoothing_factor)
        __tukey(phi_next, dx_next, smoothing_factor)
    elif(method=='guo'):
        __guo(phi, dx, smoothing_factor)
        __guo(phi_next, dx_next, smoothing_factor)
    elif(method=='weikert'):
        __weikert(phi, dx, smoothing_factor)
        __weikert(phi_next, dx_next, smoothing_factor)

    DT = smoothing_factor
    
    #--- Now build the linear system ---
    
    # We begin with the internal points 
    cdef double theta = 1.0
    for i in range(1, num_points-1):
        gammaP = 2*DT/(dt[i]*(dt[i-1]+dt[i]))
        gammaM = 2*DT/(dt[i-1]*(dt[i-1]+dt[i]))
        Adia[i]  = (1.0 + gammaP*theta*phi_next[i] + gammaM*theta*phi_next[i-1])
        Adiap[i] = -theta*gammaP*phi_next[i]
        Adiam[i] = -theta*gammaM*phi_next[i-1]
        b[i] = X[i] + (1-theta)*(gammaP*phi[i]*(X[i+1]-X[i]) - gammaM*phi[i-1]*(X[i]-X[i-1]))

    # Then the boundary
    Adia[0] = 1.0
    Adiap[0]= -1.0
    b[0]    = 0.0
    Adia[num_points-1] = 1.0
    Adiam[num_points-1] = -1.0
    b[num_points-1] = 0.0
     
    # Next, we solve the linear system
    __solve_tri_linear(Adiam, Adia, Adiap,  b, Y)

    return Y



def perona_malik(X, t=None, smoothing_factor=1, gradient_threshold=1, method='classic', integration='implicit'):
    """
    perona_malik(X, t=None, smoothing_factor=1, gradient_threshold=1, method='classic', integration='implicit')
    
    Perona_Malik edge preserving smoothing of a sequence or timeseries (1D) with various gradient stop functions.
       
    Args:
        :X (ndarray):                    input timeseries.

        :t (ndarray):                    input timestamps (also dictates smoothing rate, hence Epoch in ms will smooth very fast).

        :smoothing_factor (double):      The amount of smoothing (corresponds to the time of integration).

        :gradient_threshold (double):    Dictates the order of influcial gradients (no smoothing for large gradients).

        :integration (str):              explicit or implicit integration (implicit is always stable but has slightly more diffusion).
        
        :method: 
            :'classical':    phi(z) = 1/(1+(z/K)**2).

            :'exponential':  phi(z) = exp(-np.abs(z)**2/K).

            :'tukey':        phi(z) = 0.5*(1.0 - z/K)**2)**2 if z<K and 0 otherwise.

            :'guo':          beta = 1.0 + np.abs(z/K)**2.
                            alpha = 2 - 2/beta.
                            phi(z) = 1.0 + (z/K)**alpha.
            :'weikert':      phi[i] = 1.0 - exp(-3.31488*K**8/z**8).
    Returns:
         :Y: the smoothed signal
    """
    if(integration=='implicit'):
        return __perona_malik_midpoint(X, t, smoothing_factor, gradient_threshold, method)
    else:
        return __perona_malik_explicit(X, t, smoothing_factor, gradient_threshold, method)


# cython: profile=False
import math

import numpy as np
from tslearn.metrics import dtw

cimport cython
from libc.math cimport abs, sqrt

@cython.profile(False)
cdef inline double dist(double a, double b):
    cdef double diff = a - b
    return diff * diff

@cython.profile(False)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline double array_min(double[::1] array):
    cdef double min_value = array[0]
    cdef int i
    for i in range(1, array.shape[0]):
        min_value = min_value if min_value < array[i] else array[i]
    return min_value

@cython.profile(False)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline double array_max(double[::1] array):
    cdef double max_value = array[0]
    cdef int i
    for i in range(1, array.shape[0]):
        max_value = max_value if max_value > array[i] else array[i]
    return max_value

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef whiten_multivariate(double[:,:] window_signal):
    """
    Perform whitening on signal window - it should be local to a sliding window
    """
    cdef double[:,:] s = window_signal
    result = np.empty(shape=(s.shape[1], s.shape[0]), dtype=np.float64)
    cdef double[::1,:] result_v = result.T

    cdef double mean, diff, var, std
    cdef int i, j
    for j in range(s.shape[1]):
        mean = 0
        for i in range(s.shape[0]):
            mean += s[i,j]
        mean /= s.shape[0]

        var = 0
        for i in range(s.shape[0]):
            diff = abs(mean - s[i,j])
            var += diff * diff
        var /= s.shape[0]

        std = sqrt(var)

        for i in range(s.shape[0]):
            result_v[i,j] = (s[i,j] - mean) / (std + 1e-10)

    return result.transpose()


def is_variance_ok(double[:,:] sequence, variance_thresholds):
    cdef double[:,:] s = sequence
    cdef double lower_threshold
    cdef double upper_threshold

    cdef double mean, diff, var
    cdef int i, j
    for j in range(s.shape[1]):
        mean = 0
        for i in range(s.shape[0]):
            mean += s[i,j]
        mean /= s.shape[0]

        var = 0
        for i in range(s.shape[0]):
            diff = abs(mean - s[i,j])
            var += diff * diff
        var /= s.shape[0]

        lower_threshold = variance_thresholds["lower"][j]
        upper_threshold = variance_thresholds["upper"][j]
        for i in range(s.shape[0]):
            if var < lower_threshold or var > upper_threshold:
                return False

    return True


class DTWModel(object):
    def __init__(self, window, stride, query_len, sakoeChibaWidth=0.2):
        self.window = window
        self.stride = stride
        self.sakoeChibaWidth = sakoeChibaWidth
        self.idx = self.make_envelope_index_keogh(
            query_len, window, max(1, self.sakoeChibaWidth * max(query_len, window))
        )
        self.idx = np.array(self.idx)

    def sliding_window_index(self, signal_length):
        """
        Takes length of signal and returns list of indices, each of which
        defines a sliding window
        """
        start = 0
        while (start + self.window) <= signal_length:
            yield slice(start, start + self.window)
            start += self.stride

    @staticmethod
    def distance_static(seq1, seq2, sakoeChibaWidth):
        seq1 = whiten_multivariate(seq1).transpose()
        seq2 = whiten_multivariate(seq2).transpose()
        s = 0
        if sakoeChibaWidth:
            for i in range(len(seq1)):
                dist = dtw(
                    seq1[i],
                    seq2[i],
                    global_constraint="sakoe_chiba",
                    sakoe_chiba_radius=max(1, sakoeChibaWidth * max(len(seq1[i]), len(seq2[i]))),
                )
                s += dist
        else:
            for i in range(len(seq1)):
                dist = dtw(seq1[i], seq2[i], global_constraint=None)
                s += dist
        return s

    def lb_keogh_dist(self, seq1_raw, seq2_raw, double[:,:,::1] envelope):  # seq1 is query
        cdef double[:,::1] seq1 = whiten_multivariate(seq1_raw).transpose()
        cdef double[:,::1] seq2 = whiten_multivariate(seq2_raw).transpose()
        cdef long[:,::1] idx = self.idx
        cdef double distance = 0.0
        cdef int i, j
        cdef double score1, score2, min_value, max_value
        for i in range(seq1.shape[0]):
            score1 = 0
            score2 = 0
            for j in range(seq2.shape[1]):
                if seq2[i][j] < envelope[i][j][1]:
                    score1 += dist(seq2[i,j], envelope[i,j,1])
                elif seq2[i][j] > envelope[i][j][0]:
                    score1 += dist(seq2[i][j], envelope[i][j][0])
            for j in range(seq1.shape[1]):
                max_value = array_max(seq2[i][idx[j,0] : idx[j,1]])
                min_value = array_min(seq2[i][idx[j,0] : idx[j,1]])
                if seq1[i,j] < min_value:
                    score2 += dist(seq1[i,j], min_value)
                if seq1[i,j] > max_value:
                    score2 += dist(seq1[i,j], max_value)
            distance += max(sqrt(score1), sqrt(score2))

        return distance

    def make_envelope_keogh(self, seq, sz2, radius):
        ix = self.make_envelope_index_keogh(sz2, len(seq), radius)
        res = []
        for item in ix:
            res.append((max(seq[item[0] : item[1]]), min(seq[item[0] : item[1]])))
        return res

    @staticmethod
    def make_envelope_index_keogh(sz1, sz2, radius):
        res = []
        if sz1 == 1:
            return [(0, sz2)]
        ratio = (sz2 - 1) / (sz1 - 1)
        for i in range(sz1):
            res.append((max(math.ceil(i * ratio - radius), 0), math.floor(i * ratio + radius) + 1))

        return res




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys


sys.path.append('../../')
import numpy as np
from cognite_ml.filters import perona_malik


if __name__ == "__main__":    
    X = np.random.rand(1000)
    is_ok = 0
    try:
        for method in ['classic', 'exponential', 'tukey', 'guo','weikert']:
            for integration in ['explicit', 'implicit']:
                perona_malik(X, method=method, gradient_threshold=20, smoothing_factor=50, integration=integration)
        is_ok =1
    except:
        is_ok =0

    # Check the results 
    assert(is_ok==1)


import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_equal

from cognite_ml.filters import RollingMeanSmoother

inp = pd.Series([1, 2, 3, 4, 3, 2, 1])

class TestClassRollingMean:
    def test_even(self):
        expected = pd.Series([5/4, 7/4, 10/4, 12/4, 12/4, 10/4, 7/4])
        assert_array_equal(RollingMeanSmoother(window=4).transform(inp), expected)

    def test_odd(self):
        expected = pd.Series([4/3, 6/3, 9/3, 10/3, 9/3, 6/3, 4/3])
        assert_array_equal(RollingMeanSmoother(window=3).transform(inp), expected)

    def test_fit(self):
        spikes = pd.Series([5, 5, 10, 5, 5, 5, 10, 5, 5])
        rs1 = RollingMeanSmoother().fit(spikes, max_mean_err=0)
        rs2 = RollingMeanSmoother().fit(spikes, max_mean_err=5)
        rs3 = RollingMeanSmoother().fit(spikes)
        assert rs1.window == 1
        assert rs2.window == 8
        assert rs3.window == 2

    def test_lag(self):
        expected = pd.Series([3/3, 4/3, 6/3, 9/3, 10/3, 9/3, 6/3])
        rs = RollingMeanSmoother(window=3, center=False)
        assert_array_equal(rs.transform(inp), expected)

    def test_nofit(self):
        rs = RollingMeanSmoother(window=42)
        assert rs.fit(inp).window == 42

    def test_fit_edge(self):
        mean = np.abs(inp).mean()
        rs = RollingMeanSmoother().fit(inp, max_mean_err=mean)
        assert rs.window == len(inp) - 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
import matplotlib.pyplot as plt
import numpy as np
import os 
import pandas as pd

from cognite.client import CogniteClient

from cognite_ml.filters import perona_malik


def regularity_metric(X):
    DX = X[1:]-X[:-1]
    N = len(DX)
    st = np.sum(1.0/N*np.abs(DX)**2)**0.5
    tv = 1.0/N*np.sum(np.abs(DX))

    return st, tv


def isGood(X, Y):
    Xst, Xtv = regularity_metric(X)
    Yst, Ytv = regularity_metric(Y)

    diff_st = Xst/Yst
    diff_tv = Xtv/Ytv
    
    if((np.abs(diff_st-1.0) <0.2) & (np.abs(diff_tv-1.0) >3.0)):
        res = True
    else:
        res = False

    return res
    
if __name__ == "__main__":    
    client = CogniteClient(os.environ.get("COGNITE_API_KEY"), "akerbp")
    tag = 'VAL_10_PT_81608:VALUE'
    aggregates = ['totalvariation']
    start = datetime.datetime(2018, 6, 1)
    end   = datetime.datetime(2018, 7, 1)
    data = client.datapoints.get_datapoints_frame([tag], start=start, end=end, granularity='1h', aggregates=aggregates)
    data = data.fillna(method='ffill')
    data = data.fillna(method='bfill')
    X = data[tag+'|totalvariation'].values
    is_of_quality = []
    for method in ['classic', 'exponential', 'tukey', 'guo','weikert']:
        Z = perona_malik(X, method=method, gradient_threshold=20, smoothing_factor=50)
        is_of_quality.append(isGood(X, Z))

    # Check the results 
    assert(is_of_quality == [False, True, True, False, False])



"""
The :mod:`cognite_ml.timeseries.filters` module contains various filters
for timeseries and sequences such as noise removal, smoothers, and so on.
"""

from .perona_malik import perona_malik
from .kernel_smoothing import TruncatedGaussian
from .rolling_mean import RollingMeanSmoother

from . import perona_malik
from . import kernel_smoothing
from . import rolling_mean

__all__ = ["perona_malik", "TruncatedGaussian", "RollingMeanSmoother"]



"""

@author: agleim
"""

import numpy as np
from sklearn.utils.validation import check_array
from sklearn.base import BaseEstimator, TransformerMixin


__all__ = ["ChangeDetect"]


class ChangeDetect:

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

@author: quetran
"""
from unittest import TestCase

import pandas as pd
from cognite_ml.timeseries.pattern_search import \
    pattern_search_base as searcher
from nose_parameterized import parameterized


class SimilaritySearchBaseTestCase(TestCase):
    def test_search_init(self):
        input_df = pd.DataFrame(
            [[None, 1532432275000, 3, 4], [2, 1532432285000, 3, 4]], columns=["a", "timestamp", "b", "c"]
        )
        query_df = pd.DataFrame(
            [[None, 1532432275000, 3, 4], [None, 1532432285000, 3, 4]], columns=["a", "timestamp", "b", "c"]
        )
        s = searcher.PatternSearchBase(min_range=None, max_range=None, limit=2)
        s.search_init(input_df, query_df)
        self.assertFalse(s.input_ts.isna().any().any())
        self.assertFalse(s.query_seq.isna().any().any())
        self.assertEqual(s.granularity_second, 10)

    def test_find_similar_periods(self):
        # Test invalid min/max range
        s = searcher.PatternSearchBase(min_range="10h", max_range="1h", limit=2)
        s.granularity_second = 30 * 60
        self.assertRaises(ValueError, s.find_similar_periods, None, None)

    #
    def test_combine_score_value_error(self):
        s = searcher.PatternSearchBase(min_range=None, max_range=None, limit=2)
        # Test bad result
        score1 = [{"distance": 1, "stride": 3, "len": 5, "i": 7}]
        score2 = [{"distance": 2, "stride": 5, "len": 5, "i": 7}]
        self.assertRaises(ValueError, s.combine_scores, score1, score2)

    combine_score_test_cases = [
        (
            [{"distance": 1, "stride": 3, "len": 5, "i": 9}, {"distance": 2, "stride": 4, "len": 7, "i": 7}],
            [{"distance": 3, "stride": 4, "len": 7, "i": 7}],
            [{"distance": 5, "stride": 4, "len": 7, "i": 7}],
        ),
        (
            [
                {"distance": 1, "stride": 3, "len": 5, "i": 9},
                {"distance": 2, "stride": 4, "len": 7, "i": 7},
                {"distance": 2, "stride": 4, "len": 7, "i": 8},
                {"distance": 1, "stride": 4, "len": 8, "i": 8},
            ],
            [
                {"distance": 1, "stride": 3, "len": 3, "i": 9},
                {"distance": 2, "stride": 4, "len": 7, "i": 5},
                {"distance": 3, "stride": 4, "len": 7, "i": 8},
                {"distance": 2, "stride": 4, "len": 8, "i": 8},
            ],
            [{"distance": 5, "stride": 4, "len": 7, "i": 8}, {"distance": 3, "stride": 4, "len": 8, "i": 8}],
        ),
    ]

    @parameterized.expand(combine_score_test_cases)
    def test_combine_score(self, score1, score2, expected_result):
        s = searcher.PatternSearchBase(min_range=None, max_range=None, limit=2)
        result = s.combine_scores(score1, score2)
        self.assertCountEqual(result, expected_result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

@author: quetran
"""

import math
from functools import reduce

import numpy as np
import pandas as pd


def fill_nan(dataframe, default_value=0.0):
    dataframe.fillna(method="ffill", inplace=True)
    dataframe.fillna(method="bfill", inplace=True)
    dataframe.fillna(value=default_value, inplace=True)
    return dataframe


def convert_time_delta_to_num_points(granularity, time_delta):
    try:
        time_in_second = pd.to_timedelta(time_delta).total_seconds()
        time_in_datapoints = int(time_in_second / granularity) + 1
        return time_in_datapoints
    except Exception as e:
        raise e


def convert_index(stride, word_indices: np.ndarray = None, ts_indices: np.ndarray = None):
    """
    if word_index is not None, convert word (sliding window) index to time series index
    otherwise convert ts_index to word_index
    """
    if word_indices is not None:
        return word_indices * stride
    elif ts_indices is not None:
        return ts_indices / stride
    else:
        raise ValueError("either word_index or ts_index needs to be specified")


def make_index_even(dataframe):
    "Creates time index with evenly spaced intervals. Adds NaN to timestamps with missing values.\n\n    Args:\n        dataframe (pandas.DataFrame):  Input dataframe.\n\n    Returns:\n        pandas.DataFrame: Input dataframe with evenly spaced intervals.\n    "
    timestamps = dataframe.timestamp.values
    start_time = timestamps[0]
    end_time = timestamps[(-1)]
    deltas = np.diff(timestamps, 1)
    delta = reduce((lambda x, y: math.gcd(x, y)), deltas)
    t_new = np.arange(start_time, (end_time + delta), delta)
    new_df = pd.DataFrame(t_new, columns=["timestamp"])
    dataframe.timestamp = dataframe.timestamp.apply(lambda x: np.int64(x))
    new_df.timestamp = new_df.timestamp.apply(lambda x: np.int64(x))
    return new_df.merge(dataframe, on="timestamp", how="outer").sort_values(by="timestamp").reset_index(drop=True)


def whiten_univariate(window_signal):
    """
    Perform whitening on signal window - it should be local to a sliding window
    """
    s = np.asarray(window_signal)
    mu, sd = np.mean(s), np.std(s)
    return (s - mu) / (sd + 1e-10)


def find_variance_thresholds_univariate(expected_seq):
    """
    Find variance thresholds to prune out noise
    """
    var = expected_seq.var()
    return {"lower": 0 if 0.01 * var < 1 else 0.01 * var, "upper": max(100 * var, 1)}


def find_variance_thresholds_multivariate(expected_seq):
    """
    Find variance thresholds to prune out noise
    """
    var = expected_seq.var(0)
    thresholds = {"lower": [], "upper": []}
    for variance in var:
        thresholds["lower"].append(0 if 0.01 * variance < 1 else 0.01 * variance)
        thresholds["upper"].append(max(100 * variance, 1))
    return thresholds


def calculate_overlap(start_old, end_old, start_new, end_new):
    # returns how much of the new pattern is already covered by the old
    if start_new >= start_old and start_new <= end_old:
        ol = min(end_old, end_new) - start_new + 1
    elif end_new > start_old and end_new <= end_old:
        ol = end_new - max(start_old, start_new) + 1
    else:
        ol = max(0, min(end_new, end_old) - max(start_new, start_old) + 1)
    return ol / (end_new - start_new + 1)


def extract_matches(sorted_distances, timestamp_list, overlap_ratio, n_top):
    matches = []

    for dict in sorted_distances:
        ts_indices = convert_index(
            dict["stride"], word_indices=np.arange(int(len(timestamp_list) / dict["stride"]) + 1)
        )
        start_pattern = int(timestamp_list[ts_indices[dict["i"]]])
        end_pattern = int(timestamp_list[ts_indices[dict["i"]] + dict["len"] - 1])
        if overlap_ratio:
            for match in matches:
                if calculate_overlap(match["from"], match["to"], start_pattern, end_pattern) > overlap_ratio:
                    break
            # Only add match if overlap with better matches is less than threshold
            else:
                matches.append({"from": start_pattern, "to": end_pattern, "score": dict["distance"]})

        else:
            matches.append({"from": start_pattern, "to": end_pattern, "score": dict["distance"]})
        if len(matches) == n_top:
            break
    return matches


def convert_scores_to_percentages(matches, max_length):
    distance = max_length * 1  # 1 std away
    for m in matches:
        m["percentage"] = (1 - min(1, round(m["score"] / distance, 2))) * 100

    return matches



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# original source code made available under the Mozilla Public License 2.0
# https://github.com/paolodedios/shift-detect

"""
Relative Unconstrained Least Squares Importance Fitting algorithm
"""
from __future__ import print_function

import numpy as np
from scipy import linalg


class GaussianKernel:
    """
    Computes the n-dimensional Gaussian kernel/RBF matrix

    K(X,Y) = exp( -(|| X - Y ||^2) / (2 * sigma^2) )

    for X the data matrix of sample points
    and Y the matrix of Gaussian centers.
    """

    sigma_width = None

    def __init__(self, sigma=1.0):
        self.sigma_width = sigma

    def compute_distance(self, samples=None, sample_means=None):
        """
        Compute the distances between points in the sample's feature space
        to points along the center of the distribution
        """
        (sample_rows, sample_cols) = samples.shape
        (mean_rows, mean_cols) = sample_means.shape

        squared_samples = sum(samples ** 2, 0)
        squared_means = sum(sample_means ** 2, 0)

        return (
            np.tile(squared_means, (sample_cols, 1))
            + np.tile(squared_samples[:, None], (1, mean_cols))
            - 2 * np.dot(samples.T, sample_means)
        )

    def apply(self, samples=None, sample_means=None):
        """
        Computes an n-dimensional Gaussian/RBF kernel matrix by taking points
        in the sample's feature space and maps them to kernel coordinates in
        Hilbert space by calculating the distance to each point in the sample
        space and taking the Gaussian function of the distances.

           K(X,Y) = exp( -(|| X - Y ||^2) / (2 * sigma^2) )

        where X is the matrix of data points in the sample space,
              Y is the matrix of gaussian centers in the sample space
             sigma is the width of the gaussian function being used
        """
        squared_distance = self.compute_distance(samples, sample_means)

        return np.exp(-squared_distance / (2 * (self.sigma_width ** 2)))


class AlphaRelativeDensityRatioEstimator:
    """
    Computes the alpha-relative density ratio estimate of P(X_ref) and P(X_test)

    The alpha-relative density ratio estimator, r_alpha(X), is given by the
    following kernel model:

    g(X; theta) = SUM( (theta_l * K(X, X_centers_l)), l=0, n )

    where theta is a vector of parameters [theta_1, theta_2, ..., theta_l]^T
    to be learned from the data samples. The parameters theta in the model
    g(X; theta) is calculated by solving the following optimization problem:

      theta_hat = argmin [ ( (1/2) * theta^T * H_hat * theta) -
                    (h_hat^T * theta) + ( lambda/2 * theta^T * theta) ]

    where the expression (lamba/2 * theta^T * theta), with lambda >= 0, is
    a regularization term used to penalize against overfitting

    Reference:
    Relative Density-Ratio Estimation for Robust Distribution Comparison. Makoto Yamada,
    Taiji Suzuki, Takafumi Kanamori, Hirotaka Hachiya, and Masashi Sugiyama. NIPS,
    page 594-602. (2011)
    """

    alpha_constraint = None
    sigma_width = None
    lambda_regularizer = None
    kernel_basis = None

    def __init__(self, alpha_constraint=0.0, sigma_width=1.0, lambda_regularizer=0.0, kernel_basis=1):
        self.alpha_constraint = alpha_constraint
        self.sigma_width = sigma_width
        self.lambda_regularizer = lambda_regularizer
        self.kernel_basis = kernel_basis

    def apply(self, reference_samples=None, test_samples=None, gaussian_centers=None):
        """
        Computes the alpha-relative density ratio, r_alpha(X), of P(X_ref) and P(X_test)

          r_alpha(X) = P(Xref) / (alpha * P(Xref) + (1 - alpha) * P(X_test)

        Returns density ratio estimate at X_ref, r_alpha_ref, and at X_test, r_alpha_test
        """
        # Apply the kernel function to the reference and test samples
        K_ref = GaussianKernel(self.sigma_width).apply(reference_samples, gaussian_centers).T
        K_test = GaussianKernel(self.sigma_width).apply(test_samples, gaussian_centers).T

        # Compute the parameters, theta_hat, of the density ratio estimator
        H_hat = AlphaRelativeDensityRatioEstimator.H_hat(self.alpha_constraint, K_ref, K_test)
        h_hat = AlphaRelativeDensityRatioEstimator.h_hat(K_ref)
        theta_hat = AlphaRelativeDensityRatioEstimator.theta_hat(
            H_hat, h_hat, self.lambda_regularizer, self.kernel_basis
        )

        # Estimate the density ratio, r_alpha_ref = r_alpha(X_ref)
        r_alpha_ref = AlphaRelativeDensityRatioEstimator.g_of_X_theta(K_ref, theta_hat).T
        # Estimate the density ratio, r_alpha_test = r_alpha(X_test)
        r_alpha_test = AlphaRelativeDensityRatioEstimator.g_of_X_theta(K_test, theta_hat).T

        return (r_alpha_ref, r_alpha_test)

    @staticmethod
    def H_hat(alpha=0.0, kernel_matrix_reference_samples=None, kerne_matrix_test_samples=None):
        """
        Calculates the H_hat term of the theta_hat
        optimization problem.
        """
        N_ref = kernel_matrix_reference_samples.shape[1]
        N_test = kerne_matrix_test_samples.shape[1]

        H_hat = (alpha / N_ref) * np.dot(kernel_matrix_reference_samples, kernel_matrix_reference_samples.T) + (
            (1.0 - alpha) / N_test
        ) * np.dot(kerne_matrix_test_samples, kerne_matrix_test_samples.T)

        return H_hat

    @staticmethod
    def h_hat(kernel_matrix_reference_samples):
        """
        Calculates the h_hat term of the theta_hat optimization problem
        """
        h_hat = np.mean(kernel_matrix_reference_samples, 1)

        return h_hat

    @staticmethod
    def theta_hat(H_hat=None, h_hat=None, lambda_regularizer=0.0, kernel_basis=None):
        """
        Calculates theta_hat given H_hat, h_hat, lambda, and the kernel basis function
        Treat as a system of lienar equations and find the exact, optimal
        solution
        """
        theta_hat = linalg.solve(H_hat + (lambda_regularizer * np.eye(kernel_basis)), h_hat)

        return theta_hat

    @staticmethod
    def J_of_theta(alpha=0.0, g_X_ref_theta=None, g_X_test_theta=None):
        """
        Calculates the squared error criterion, J
        """
        return (
            (alpha / 2.0) * (np.mean(g_X_ref_theta ** 2))
            + ((1 - alpha) / 2.0) * (np.mean(g_X_test_theta ** 2))
            - np.mean(g_X_ref_theta)
        )

    @staticmethod
    def g_of_X_theta(kernel_matrix_samples=None, theta_hat=None):
        """
        Calculate the alpha-relative density ratio kernel model
        """
        return np.dot(kernel_matrix_samples.T, theta_hat)


class PearsonRelativeDivergenceEstimator:
    """
    Calculates the alpha-relative Pearson divergence score

    The alpha-relative Pearson divergence is given by the following expression:

      PE_alpha = -(alpha/2(n_ref)) * SUM(r_alpha(X_ref_i)^2, i=0, n_ref)        -
                  ((1-alpha)/2(n_test)) * SUM(r_alpha(X_test_j)^2, j=0, n_test) +
                  (1/n_ref) * SUM(r_alpha(X_ref_i), i=0, n_ref)                 -
                  1/2

    where r_alpha(X) is the alpha-relative density ratio estimator and is given by
    the following kernel model:

      g(X; theta) = SUM( (theta_l * K(X, X_centers_l)), l=0, n )

    Reference:
    Relative Density-Ratio Estimation for Robust Distribution Comparison. Makoto
    Yamada, Taiji Suzuki, Takafumi Kanamori, Hirotaka Hachiya, and Masashi Sugiyama.
    NIPS, page 594-602. (2011)
    """

    alpha_constraint = None
    sigma_width = None
    lambda_regularizer = None
    kernel_basis = None

    def __init__(self, alpha_constraint=0.0, sigma_width=1.0, lambda_regularizer=0.0, kernel_basis=1):
        self.alpha_constraint = alpha_constraint
        self.sigma_width = sigma_width
        self.lambda_regularizer = lambda_regularizer
        self.kernel_basis = kernel_basis

    def apply(self, reference_samples=None, test_samples=None, gaussian_centers=None):
        """
        Calculates the alpha-relative Pearson divergence score
        """
        density_ratio_estimator = AlphaRelativeDensityRatioEstimator(
            self.alpha_constraint, self.sigma_width, self.lambda_regularizer, self.kernel_basis
        )

        # Estimate alpha relative density ratio and pearson divergence score
        (r_alpha_X_ref, r_alpha_X_test) = density_ratio_estimator.apply(
            reference_samples, test_samples, gaussian_centers
        )

        PE_divergence = (
            np.mean(r_alpha_X_ref)
            - (
                0.5
                * (
                    self.alpha_constraint * np.mean(r_alpha_X_ref ** 2)
                    + (1.0 - self.alpha_constraint) * np.mean(r_alpha_X_test ** 2)
                )
            )
            - 0.5
        )

        return (PE_divergence, r_alpha_X_test)


class RULSIF:
    """
    Estimates the alpha-relative Pearson Divergence via Least Squares Relative
    Density Ratio Approximation

    Reference:
    Relative Density-Ratio Estimation for Robust Distribution Comparison. Makoto
    Yamada, Taiji Suzuki, Takafumi Kanamori, Hirotaka Hachiya, and Masashi Sugiyama.
    NIPS, page 594-602. (2011)
    """

    alpha_constraint = None
    sigma_width = None
    lambda_regularizer = None
    kernel_basis = None
    cross_folds = None
    gaussian_centers = None

    def __init__(self, alpha_constraint=0.0, sigma_width=1.0, lambda_regularizer=0.0, kernel_basis=100, cross_folds=5):

        self.alpha_constraint = alpha_constraint
        self.sigma_width = sigma_width
        self.lambda_regularizer = lambda_regularizer
        self.kernel_basis = kernel_basis
        self.cross_folds = cross_folds
        self.gaussian_centers = None

    def get_median_distance_between_samples(self, sample_set=None):
        """
        Jaakkola's heuristic method for setting the width parameter of the Gaussian
        radial basis function kernel is to pick a quantile (usually the median) of
        the distribution of Euclidean distances between points having different
        labels.

        Reference:
        Jaakkola, M. Diekhaus, and D. Haussler. Using the Fisher kernel method to detect
        remote protein homologies. In T. Lengauer, R. Schneider, P. Bork, D. Brutlad, J.
        Glasgow, H.- W. Mewes, and R. Zimmer, editors, Proceedings of the Seventh
        International Conference on Intelligent Systems for Molecular Biology.
        """
        numrows = sample_set.shape[0]
        samples = sample_set

        G = np.sum((samples * samples), 1)
        Q = np.tile(G[:, np.newaxis], (1, numrows))
        R = np.tile(G, (numrows, 1))

        distances = Q + R - 2 * np.dot(samples, samples.T)
        distances = distances - np.tril(distances)
        distances = distances.reshape(numrows ** 2, 1, order="F").copy()

        return np.sqrt(0.5 * np.median(distances[distances > 0]))

    def compute_gaussian_width_candidates(self, reference_samples=None, test_samples=None):
        """
        Compute a candidate list of Gaussian kernel widths. The best width will be
        selected via cross-validation
        """
        all_samples = np.c_[reference_samples, test_samples]
        median_distance = self.get_median_distance_between_samples(all_samples.T)

        return median_distance * np.array([0.6, 0.8, 1, 1.2, 1.4])

    def generate_regularization_params(self):
        """
        Generatees a candidate list of regularization parameters to be used
        with the L1 regularizer term of RULSIF optimization problem.  The
        best regularizer parameter will be chosen via cross-validation
        """
        return 10.0 ** np.array([-3, -2, -1, 0, 1])

    def generate_all_gaussian_centers(self, reference_samples=None):
        """
        Generates kernels in the region where the P(X_reference) takes large values
        """
        self.kernel_basis = reference_samples.shape[1]

        return reference_samples[:, np.r_[0 : self.kernel_basis]]

    def generate_random_gaussian_centers(self, reference_samples=None):
        """
        Randomly chooses Gaussian centers as an optimization
        """
        numcols = reference_samples.shape[1]
        reference_sample_idxs = np.random.permutation(numcols)
        self.kernel_basis = min(self.kernel_basis, numcols)

        return reference_samples[:, reference_sample_idxs[0 : self.kernel_basis]]

    def generate_first_N_gaussian_centers(self, reference_samples=None):
        """
        Chooses the firts N samples as Gaussian centers as an optimization
        """
        numcols = reference_samples.shape[1]
        self.kernel_basis = min(self.kernel_basis, numcols)

        return reference_samples[:, np.r_[0 : self.kernel_basis]]

    def generate_gaussian_centers(self, reference_samples=None):
        """
        Choose Gaussian centers based on a strategy
        """
        # gaussianCenters = self.generateAllGaussianCenters(referenceSamples)
        gaussian_centers = self.generate_random_gaussian_centers(reference_samples)

        return gaussian_centers

    def compute_model_parameters(self, reference_samples=None, test_samples=None, gaussian_centers=None):
        """
        Computes model parameters via k-fold cross validation process
        """
        (ref_rows, ref_cols) = reference_samples.shape
        (test_rows, test_cols) = test_samples.shape

        sigma_widths = self.compute_gaussian_width_candidates(reference_samples, test_samples)
        lambda_candidates = self.generate_regularization_params()

        # Initialize cross validation scoring matrix
        cross_validation_scores = np.zeros((np.size(sigma_widths), np.size(lambda_candidates)))

        # Initialize a cross validation index assignment list
        reference_samples_cv_idxs = np.random.permutation(ref_cols)
        reference_samples_cv_split = np.floor(np.r_[0:ref_cols] * self.cross_folds / ref_cols)
        test_samples_cv_idxs = np.random.permutation(test_cols)
        test_samples_cv_split = np.floor(np.r_[0:test_cols] * self.cross_folds / test_cols)

        # Initiate k-fold cross-validation procedure. Using variable
        # notation similar to the RULSIF formulas.
        for sigma_idx in np.r_[0 : np.size(sigma_widths)]:

            # (re-)Calculate the kernel matrix using the candidate sigma width
            sigma = sigma_widths[sigma_idx]
            K_ref = GaussianKernel(sigma).apply(reference_samples, gaussian_centers).T
            K_test = GaussianKernel(sigma).apply(test_samples, gaussian_centers).T

            # Initialize a new result matrix for the current sigma candidate
            fold_result = np.zeros((self.cross_folds, np.size(lambda_candidates)))

            for fold_idx in np.r_[0 : self.cross_folds]:

                K_ref_training_set = K_ref[:, reference_samples_cv_idxs[reference_samples_cv_split != fold_idx]]
                K_test_training_set = K_test[:, test_samples_cv_idxs[test_samples_cv_split != fold_idx]]

                H_h_kth_fold = AlphaRelativeDensityRatioEstimator.H_hat(
                    self.alpha_constraint, K_ref_training_set, K_test_training_set
                )
                h_h_kth_fold = AlphaRelativeDensityRatioEstimator.h_hat(K_ref_training_set)

                for lambda_idx in np.r_[0 : np.size(lambda_candidates)]:
                    lambda_candidate = lambda_candidates[lambda_idx]

                    theta_h_kth_fold = AlphaRelativeDensityRatioEstimator.theta_hat(
                        H_h_kth_fold, h_h_kth_fold, lambda_candidate, self.kernel_basis
                    )

                    # Select the subset of the kernel matrix not used in the training set
                    # for use as the test set to validate against
                    k_ref_test_set = K_ref[:, reference_samples_cv_idxs[reference_samples_cv_split == fold_idx]]
                    k_test_test_set = K_test[:, test_samples_cv_idxs[test_samples_cv_split == fold_idx]]

                    r_alpha_X_ref = AlphaRelativeDensityRatioEstimator.g_of_X_theta(k_ref_test_set, theta_h_kth_fold)
                    r_alpha_X_test = AlphaRelativeDensityRatioEstimator.g_of_X_theta(k_test_test_set, theta_h_kth_fold)

                    # Calculate the objective function J(theta) under the current parameters
                    J = AlphaRelativeDensityRatioEstimator.J_of_theta(
                        self.alpha_constraint, r_alpha_X_ref, r_alpha_X_test
                    )

                    fold_result[fold_idx, lambda_idx] = J

                cross_validation_scores[sigma_idx, :] = np.mean(fold_result, 0)

        cross_validation_min_scores = cross_validation_scores.min(1)
        cross_validation_min_idx_for_lambda = cross_validation_scores.argmin(1)
        cross_validation_min_idx_for_sigma = cross_validation_min_scores.argmin()

        optimal_sigma = sigma_widths[cross_validation_min_idx_for_sigma]
        optimal_lambda = lambda_candidates[cross_validation_min_idx_for_lambda[cross_validation_min_idx_for_sigma]]

        return (optimal_sigma, optimal_lambda)

    def fit(self, reference_samples=None, test_samples=None):
        """
        Learn the proper model parameters
        """

        # Reset RNG to ensure consistency of experimental results.  In a production
        # environment, the RNG should use a truly random seed and hyper-parameters
        np.random.seed(0)

        gaussian_centers = self.generate_gaussian_centers(reference_samples)

        (optimal_sigma, optimal_lambda) = self.compute_model_parameters(
            reference_samples, test_samples, gaussian_centers
        )

        self.sigma_width = optimal_sigma
        self.lambda_regularizer = optimal_lambda
        self.gaussian_centers = gaussian_centers

    def predict(self, reference_samples=None, test_samples=None):
        """
        Estimates the alpha-relative Pearson divergence as determined by the relative
        ratio of probability densities:

           P(ReferenceSamples[x]) / (alpha * P(ReferenceSamples[x]) + (1 - alpha) * P(TestSamples[x]))

        from samples:
           ReferenceSamples[x_i] | ReferenceSamples[x_i] in R^{d}, with i=1 to ReferenceSamples{N}

        drawn independently from P(ReferenceSamples[x])

        and from samples:
           TestSamples[x_j] | TestSamples[x_j] in R^{d}, with j=1 to TestSamples{N}

        drawn independently from P(TestSamples[x])

        After the model hyper-parameters have been learned and chosen by the train()
        method, the RULSIF algorithm can be applied repeatedly on both in-sample and out
        of sample data
        """

        if self.gaussian_centers is None or self.kernel_basis is None:
            raise Exception("Missing kernel basis function parameters")

        if self.sigma_width == 0.0 or self.lambda_regularizer == 0.0:
            raise Exception("Missing model selection parameters")

        divergence_estimator = PearsonRelativeDivergenceEstimator(
            self.alpha_constraint, self.sigma_width, self.lambda_regularizer, self.kernel_basis
        )
        (PE_alpha, r_alpha_X_test) = divergence_estimator.apply(reference_samples, test_samples, self.gaussian_centers)

        return PE_alpha



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import math
import pandas as pd
from cognite.client._utils import granularity_to_ms

__all__ = ["RangeBasedPrecisionRecall"]


class RangeBasedPrecisionRecall:
    """Class for Precision and Recall for Range-Based Anomaly Detection

    :param anomaly_labels: list of timestamp tuples of actual anomaly labels with (start_time, end_time)
    :param predicted_labels: list of timestamp tuples of predicted anomaly labels with (start_time, end_time)
    :param alpha: relative weight of existence reward
    :param beta: relative weight of overlap reward
    :param bias_type: positional bias type ('front', 'flat', 'tail')
    :param granularity: granularity of tieme series data used to generate the predicted anomaly labels

    Lee et al. 2018
    https://arxiv.org/pdf/1801.03175.pdf
    """

    def __init__(self, anomaly_labels, predicted_labels, alpha=0.5, beta=0.5, bias_type=None, granularity=None):

        self.anomaly_labels = anomaly_labels
        self.predicted_labels = predicted_labels
        self.alpha = alpha
        self.beta = beta
        self.bias_type = bias_type
        self.granularity = granularity
        self.precision = None
        self.recall = None

    def __is_contained(self, a_range, b_range):
        """returns whether a_range is contained in b_range where range is given by a tuple of (start, end) timestamps"""
        if a_range[0] >= b_range[0] and a_range[1] <= b_range[1]:
            return True
        else:
            return False

    def __is_empty(self, any_structure):
        """returns whether a data structure is empty or not"""
        if any_structure:
            return False
        else:
            return True

    def __get_intersection(self, a_range, b_range):
        """returns the union of two ranges a and b where each range is given by a tuple of (start, end) timestamps"""
        a_left, a_right = a_range
        b_left, b_right = b_range
        end_pts = sorted([a_left, a_right, b_left, b_right])
        middle = [end_pts[1], end_pts[2]]
        if a_right >= b_left and b_right >= a_left:
            intersection = middle
        else:
            intersection = []

        return tuple(intersection)

    def __get_length(self, a_range):
        """returns the length of a range a given by a tuple of (start, end)"""
        length = a_range[1] - a_range[0]  # measured in 'ms'

        if self.granularity is not None:
            granularity_in_ms = granularity_to_ms(self.granularity)
            length = length / granularity_in_ms

        return length

    def __get_step_ranges(self, a_range):
        """returns a list of timestamps that span the range given by a tuple of (start, end) times"""
        if self.granularity is not None:
            step = granularity_to_ms(self.granularity)
        else:
            step = 1

        points = list(range(a_range[0], a_range[1] + step, step))

        return [(points[i], points[i + 1]) for i in range(len(points) - 1)]

    def __positional_bias(self, position, range_length):
        """returns positional bias based on type"""
        if self.bias_type == "front":
            return range_length - position + 1
        elif self.bias_type == "tail":
            return position
        else:
            return 1

    def __overlap_size(self, a_range, b_range):
        """returns the relative overlap size of a_range wrt a_range \intersect b_range"""
        my_value = 0
        max_value = 0
        step_ranges = self.__get_step_ranges(a_range)
        intersection = self.__get_intersection(a_range, b_range)
        range_length = len(step_ranges)

        if self.__is_empty(intersection) or range_length == 0:
            return 0.0
        else:
            for i in range(range_length):
                bias = self.__positional_bias(i, range_length)
                max_value = max_value + bias
                step_range = step_ranges[i]
                if self.__is_contained(step_range, intersection):
                    my_value = my_value + bias

            return my_value / max_value

    def __cardinality_factor(self, a_range, list_range):
        """returns a factor that is inversely proportional to the cardinality of anomaly_range.
        Takes the value 1 if anomaly_range overlaps with at most one predicted range in predicted_labels"""
        overlaps = []
        for b_range in list_range:
            overlap = self.__get_intersection(a_range, b_range)
            if not self.__is_empty(overlap):
                overlaps.append(overlap)

        if len(overlaps) >= 1:
            return 1 / len(overlaps)
        else:
            return 0.0

    def __overlap_reward(self, anomaly_range, predicted_labels):
        """returns the overlap reward based on size, position and cardinality
        of one anomaly range wrt all predicted anomaly labels"""

        return self.__cardinality_factor(anomaly_range, predicted_labels) * sum(
            [self.__overlap_size(anomaly_range, predicted_range) for predicted_range in predicted_labels]
        )

    def __existence_reward(self, a_range, list_range):
        """returns the existence reward of one anomaly_range wrt all predicted anomaly labels"""

        overlaps = []
        for b_range in list_range:
            overlap = self.__get_intersection(a_range, b_range)
            if not self.__is_empty(overlap):
                overlaps.append(overlap)

        if len(overlaps) >= 1:
            return 1
        else:
            return 0.0

    def __recall_of_one_range(self, anomaly_range, predicted_labels):
        """returns recall score of one anomaly range wrt all predicted anomaly labels"""
        return self.alpha * self.__existence_reward(
            anomaly_range, predicted_labels
        ) + self.beta * self.__overlap_reward(anomaly_range, predicted_labels)

    def __precision_of_one_range(self, prediction_range, anomaly_labels):
        """returns the precision based on size, position and cardinality
        of one predicted anomaly range wrt all anomaly labels"""

        return self.__cardinality_factor(prediction_range, anomaly_labels) * sum(
            [self.__overlap_size(prediction_range, anomaly_range) for anomaly_range in anomaly_labels]
        )

    def get_recall(self):
        """returns recall score of all anomaly labels wrt all predicted anomaly labels"""
        vals = [
            self.__recall_of_one_range(anomaly_range, self.predicted_labels) for anomaly_range in self.anomaly_labels
        ]

        self.recall = sum(vals) / len(vals)

        return self.recall

    def get_precision(self):
        """returns precision score of all anomaly labels wrt all predicted anomaly labels"""
        vals = [
            self.__precision_of_one_range(prediction_range, self.anomaly_labels)
            for prediction_range in self.predicted_labels
        ]

        self.precision = sum(vals) / len(vals)

        return self.precision

    def get_detection_lags(self):
        """returns the detection lag of first flagged anomaly range on or after each anomaly start"""
        detect_lags = []
        anom_lengths = []

        for i in range(len(self.anomaly_labels)):
            an_win = self.anomaly_labels[i]
            next_an_start = self.anomaly_labels[i + 1][0] if i < len(self.anomaly_labels) - 1 else math.inf
            try:
                first_detect = sorted(
                    [pr_win[0] for pr_win in self.predicted_labels if an_win[0] <= pr_win[0] < next_an_start]
                )[0]
                detect_lags.append(pd.to_datetime(first_detect, unit="ms") - pd.to_datetime(an_win[0], unit="ms"))
            except IndexError:
                detect_lags.append("Not detected")
            anom_lengths.append(pd.to_datetime(an_win[1], unit="ms") - pd.to_datetime(an_win[0], unit="ms"))

        return detect_lags, anom_lengths

    def print_test_performance(self):
        """ returns nothing: prints summary of algorithm performance on test case"""
        print("Range-based Precision:", round(100 * self.get_precision(), 2), "%")
        print("Range-based    Recall:", round(100 * self.get_recall(), 2), "%")
        anomaly_detection_lags, anomaly_window_lengths = self.get_detection_lags()
        print("Delay(s) before first detection of anomaly:")
        for i in range(len(anomaly_detection_lags)):
            if anomaly_detection_lags[i] == "Not detected":
                print("  #{}: Not detected".format(i + 1))
            else:
                print("  #{}: {} into window of {}".format(i + 1, anomaly_detection_lags[i], anomaly_window_lengths[i]))



from functools import partial
from multiprocessing import Pool, cpu_count

import numpy as np
from cognite_ml.timeseries.pattern_search import auxiliary as aux


class PatternSearchBase:
    def __init__(self, min_range, max_range, limit):
        self.limit = limit
        self.min_range = min_range
        self.max_range = max_range
        self.overlap_th = 0.1
        self.filterVariance = True

        self.input_ts = None
        self.query_seq = None
        self.granularity_second = None

    def search_init(self, input_timeseries, pattern_timeseries, timeout=None):
        self.timeout = timeout

        self.input_ts = input_timeseries
        self.query_seq = pattern_timeseries

        # make data evenly spaced
        self.input_ts = aux.make_index_even(self.input_ts)
        self.query_seq = aux.make_index_even(self.query_seq)

        # fill in missing values again in case there still are some NaNs
        aux.fill_nan(self.input_ts)
        aux.fill_nan(self.query_seq)

        timestamps = self.input_ts.timestamp.values
        self.granularity_second = int((timestamps[1] - timestamps[0]) / 1000)

    def find_similar_periods(self, expected_seq, timeseries):
        min_length = (
            aux.convert_time_delta_to_num_points(self.granularity_second, self.min_range)
            if self.min_range
            else len(expected_seq)
        )
        max_length = (
            aux.convert_time_delta_to_num_points(self.granularity_second, self.max_range)
            if self.max_range
            else len(expected_seq)
        )

        if max_length < min_length:
            raise ValueError("minRange is larger than maxRange.")

        lengths = sorted(set(map(int, np.linspace(min_length, max_length, 10, endpoint=True))))
        self.max_length = lengths[len(lengths) - 1]
        distances = []

        num_cpu = cpu_count()
        with Pool(processes=num_cpu) as p:
            results = p.map_async(
                partial(self.find_similar_patterns, expected_seq=expected_seq, timeseries=timeseries), lengths
            ).get(timeout=self.timeout)

        for res in results:
            distances += res
        return distances

    def combine_scores(self, score1, score2):
        """
        Combines the scores in score1 and score2 when length and index matches.
        """
        result = []
        j1 = 0
        j2 = 0

        while j1 < len(score1) and j2 < len(score2):
            len1 = score1[j1]["len"]
            len2 = score2[j2]["len"]
            i1 = score1[j1]["i"]
            i2 = score2[j2]["i"]
            stride1 = score1[j1]["stride"]
            stride2 = score2[j2]["stride"]

            if len1 == len2:
                if i1 == i2:
                    if stride1 != stride2:
                        raise ValueError("different strides")
                    result.append(
                        {
                            "distance": score1[j1]["distance"] + score2[j2]["distance"],
                            "stride": stride1,
                            "len": len1,
                            "i": i1,
                        }
                    )
                    j1 += 1
                    j2 += 1
                elif i1 < i2:
                    j1 += 1
                else:
                    j2 += 1

            elif len1 < len2:
                j1 += 1

            else:
                j2 += 1
        return result

    def find_similar_patterns(self, search_len, expected_seq, timeseries):
        raise NotImplementedError

    def search(self, dataframes):
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Filters based on calculating rolling means.
@author: Dmytro Shynkevych
"""

import numpy as np
import pandas as pd

from sklearn.base import TransformerMixin


class RollingMeanSmoother(TransformerMixin):
    """A rolling mean smoother with switchable kernel.

    Args:
        window (int, optional): Number of timestamps to include in the window.
            If not given, the smoother will have to be fitted before use.
        center: (bool, optional): Whether the window should be centered on current datum.
            If False, the window will lag behind (ending at current datum). Default is True.
        kernel: (str, optional): If given, the mean will be weighted according to the kernel.
            Otherwise, an unweighted mean will be used.
            The value must be one of the `win_type`s supported by scipy.signal.
        **kwargs: Kernel-specific keyword arguments.
    """

    def __init__(self, window=None, center=True, kernel=None, **kwargs):
        self.center = center
        self.kernel = kernel
        self._kernel_kwargs = kwargs
        self.window = window

    def fit(self, data, max_mean_err=None):
        """
        Fit the window size to given time series with given error budget.

        Args:
            data (pd.Series): The data to fit to. Index must be monotonic.
            max_mean_err (float, optional): Maximal allowable deviation from original.
                The default is 20% of the average magnitude, which is empirically reasonable.

        Returns:
            RollingMeanSmoother: A fitted smoother.
        """
        if self.window is not None:  # explicitly given or already fitted
            return self
        if max_mean_err is None:
            max_mean_err = np.abs(data).mean() * 0.2

        # Bisect on rmse of transformed data
        lo, hi = 1, data.shape[0]
        while lo < hi:
            mid = (lo + hi) // 2
            self.window = mid
            transformed = self.transform(data)
            mean_err = np.mean(np.abs(data - transformed))
            if mean_err > max_mean_err:
                hi = mid
            else:
                lo = mid + 1
        self.window = lo - 1  # lo is the least such that rmse(lo) > max

        return self

    def transform(self, data):
        """Apply smoothing to given time series.
        The series is padded in both directions with the first and the last datum
        respectively so that the mean can be calculated everywhere.

        Args:
            data (pd.Series): The data to smoothen. Index must be monotonic.

        Returns:
            pd.Series: Smoothened output of the same shape as input data.
        """
        if self.center:
            left = self.window // 2
            right = (self.window - 1) // 2
            lpad = pd.Series([data.iloc[0]] * left)
            rpad = pd.Series([data.iloc[-1]] * right)
            data = lpad.append(data).append(rpad)
        else:
            data = pd.Series([data.iloc[0]] * self.window).append(data)

        res = data.rolling(self.window, win_type=self.kernel, center=self.center).mean(**self._kernel_kwargs)

        if self.center:
            if right > 0:  # [...:-0] is not a valid range
                return res.iloc[left:-right]
            else:
                return res.iloc[left:]
        else:
            return res.iloc[self.window :]



"""

@author: Johannes Kolberg
"""

import pandas as pd


def anomaly_ranges(anomaly_flags, threshold=None):
    """Finds the time ranges with continuous values of 1.0.
    Applies a threshold if specified, e.g. if inputs are continuous.

    :param anomaly_flags: Pandas df where first col predictions, index is timestamps.
    :param (optional) threshold: Values above are set to 1, those below to 0.
    :return: List of tuples with start and end timestamps of anomalous ranges in time.
    """
    if threshold:
        anomaly_flags.iloc[:, 0] = anomaly_flags.iloc[:, 0].apply(lambda y: 1 if y >= threshold else 0)

    anom_groups = anomaly_flags.iloc[:, 0].to_frame(name="anomaly")
    anom_groups["value_grp"] = (anom_groups.anomaly.diff(1) != 0).astype("int").cumsum()
    anom_groups["timestamp"] = anom_groups.index.values

    flag_ranges = pd.DataFrame(
        {
            "Begin": anom_groups.groupby("value_grp").timestamp.first(),
            "End": anom_groups.groupby("value_grp").timestamp.last(),
            "Consecutive": anom_groups.groupby("value_grp").size(),
            "anomaly": anom_groups.groupby("value_grp").anomaly.first(),
        }
    ).reset_index(drop=True)

    anom_ranges = []
    for index, _ in flag_ranges[flag_ranges.anomaly == 1].iterrows():
        anom_ranges.append((flag_ranges.Begin[index], flag_ranges.End[index]))

    return anom_ranges



import numpy as np
import pandas as pd
from numpy.testing import assert_array_equal

from cognite_ml.transformations.cluster_span import ClusterSpanFinder

inp = pd.DataFrame({'num': [.1, .2, .3, 10, 9, 3, 3.5, 4, 4.5, .15],
                    'cluster': [1, 1, 1, 3, 3, 2, 2, 2, 2, 1]})

class TestClassClusterFilter:
    def test_find(self):
        expected = np.array([3, 3, 3, 2, 2, 4, 4, 4, 4, 1])
        csf = ClusterSpanFinder()
        assert_array_equal(csf.fit_transform(inp.cluster), expected)



import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_almost_equal as assert_close

import statsmodels.api as sm

from cognite_ml.models import SMRegressor


@pytest.fixture(scope="session")
def train():
    # True relationship: 3 * x
    return pd.DataFrame({"x": [1, 2, 4, 7, 10, 15], "y": [3, 5.5, 13.5, 21, 30, 45]})


@pytest.fixture(scope="session")
def estimator(train):
    return SMRegressor(sm.OLS, fit_intercept=True).fit(train[["x"]], train["y"])


@pytest.fixture(scope="session")
def estimator_noconst(train):
    return SMRegressor(sm.OLS, fit_intercept=False).fit(train[["x"]], train["y"])


class TestClassSMRegressor:
    def test_predict(self, train, estimator, estimator_noconst):
        assert_close(pd.Series([15]), estimator.predict(pd.DataFrame({"x": [5]})), decimal=0)
        assert_close(pd.Series([15]), estimator_noconst.predict(pd.DataFrame({"x": [5]})), decimal=0)

    def test_summary(self, estimator):
        assert estimator.summary() is not None

    def test_importance(self, estimator, estimator_noconst):
        assert_close(1 / estimator.feature_importances_, [0], decimal=2)
        assert_close(1 / estimator_noconst.feature_importances_, [0], decimal=2)

    def test_coef(self, estimator, estimator_noconst):
        assert_close(estimator.coef_, [3], decimal=2)
        assert_close(estimator_noconst.coef_, [3], decimal=2)



from sklearn.base import TransformerMixin


class ClusterSpanFinder(TransformerMixin):
    """A transformer computing spans (equivalently, sequences) in clustered data.
    """

    def fit(self, clusters):
        """No-op
        """
        return self

    def transform(self, clusters):
        """Compute spans of data clusters, i.e. the number of points in the same
        cluster not separated from the current point by points in other clusters.

        Args:
            clusters (pd.Series): a series associating rows with clusters.

        Returns:
            pd.Series: a series associating rows with spans.
        """
        # Group *adjacent* observations from the same cluster together
        adjacency = (clusters != clusters.shift()).cumsum()
        spans = clusters.groupby(adjacency).size().reset_index(drop=True)
        groups = clusters.groupby(adjacency).ngroup().to_frame(name="group")

        return groups.join(spans, on="group")[spans.name].rename("span")



from sklearn.base import BaseEstimator, RegressorMixin


class SMRegressor(BaseEstimator, RegressorMixin):
    """Wrapper around a statsmodels linear regressor such as sm.OLS or sm.RLM.
    """

    def __init__(self, model_class, fit_intercept=True):
        self.model_class = model_class
        self.fit_intercept = fit_intercept
        self._fitted = False

    def fit(self, X, y, **kwargs):
        if self.fit_intercept:
            # Do not use sm.add_constant to avoid explicit statsmodels dependency
            X = X.copy()
            X.insert(0, "const", 1)
        self._model = self.model_class(y, X)
        self._results = self._model.fit()
        self._fitted = True
        return self

    def predict(self, X, **kwargs):
        if self.fit_intercept:
            X = X.copy()
            X.insert(0, "const", 1)
        return self._results.predict(X)

    transform = predict

    def summary(self):
        return self._results.summary()

    @property
    def coef_(self):
        start_idx = 1 if self.fit_intercept else 0  # Omit coefficient of added constant
        return self._results.params[start_idx:]

    @property
    def feature_importances_(self):
        start_idx = 1 if self.fit_intercept else 0  # Omit pvalue of added constant
        return 1 / (self._results.pvalues[start_idx:] + 1e-9)  # Avoid numerical problems



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

@author: agleim
"""

import pytest
import numpy as np
from numpy.testing import assert_array_equal

from cognite_ml.timeseries import SlidingWindows


class TestClassSlidingWindow(object):
    def test_1D_data_step_1(self):
        X = np.array(np.array([1, 2, 3, 4, 5]).reshape(-1, 1))
        win = SlidingWindows(window_size=2, step_size=1).transform(X)
        assert_array_equal(win, np.array([[1, 2], [2, 3], [3, 4], [4, 5]]))

    def test_1D_data_step_3(self):
        X = np.array(np.array([1, 2, 3, 4, 5]).reshape(-1, 1))
        win = SlidingWindows(window_size=2, step_size=3).transform(X)
        assert_array_equal(win, np.array([[1, 2], [4, 5]]))

    def test_1D_data_too_large_windowsize(self):
        X = np.array(np.array([1, 2, 3, 4, 5]).reshape(-1, 1))
        with pytest.raises(AssertionError):
            win = SlidingWindows(window_size=7, step_size=1).transform(X)

    def test_2D_data_step_1(self):
        X = np.array([[1, 10, 100], [2, 20, 200], [3, 30, 300]])
        win = SlidingWindows(window_size=2, step_size=1).transform(X)
        assert_array_equal(win, np.array([[1., 10., 100., 2., 20., 200.], [2., 20., 200., 3., 30., 300.]]))



import re, numpy

from setuptools import find_packages, setup
from distutils.extension import Extension
from Cython.Distutils import build_ext

packages = find_packages(exclude=["tests*"])

version = re.search('^__version__\s*=\s*"(.*)"', open("cognite_ml/__init__.py").read(), re.M).group(1)
cmdclass = { }
ext_modules = [ ]

ext_modules += [
    Extension(
        "cognite_ml.filters.perona_malik",
        ["cognite_ml/filters/perona_malik.pyx"],
        cython_directives={"c_string_type": "str", "c_string_encoding": "utf8"},
        language="c++",
    ),
    Extension(
        "cognite_ml.timeseries.pattern_search.algorithms.DTW.pydtw",
        ["cognite_ml/timeseries/pattern_search/algorithms/DTW/pydtw.pyx"],
    ),
]
cmdclass.update({"build_ext": build_ext})

setup(
    name="cognite-ml",
    version=version,
    description="Cognite Machine Learning Toolkit",
    url="https://github.com/cognitedata/cognite_ml",
    download_url="https://github.com/cognitedata/cognite_ml/archive/{}.tar.gz".format(version),
    author="Data Science Cognite",
    author_email="que.tran@cognite.com",
    packages=packages,
    install_requires=[],
    include_package_data=True,
    ext_modules = ext_modules,
    include_dirs=[numpy.get_include()],
    cmdclass = cmdclass
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

@author: quetran
"""
import unittest

import numpy as np
import pandas as pd
from cognite_ml.timeseries.pattern_search.algorithms.DTW.pydtw import (DTWModel,
                                                                       is_variance_ok,
                                                                       whiten_multivariate)


class DTWTestCase(unittest.TestCase):
    def test_is_variance_ok(self):
        sequence = pd.DataFrame([[2, 2, 3], [2, 3, 5], [2, 4, 7], [2, 5, 9]], dtype=np.float64)
        # Good variance
        variance_thresholds = {"lower": [0] * 4, "upper": [10] * 4}
        lol = sequence.values
        self.assertTrue(is_variance_ok(sequence.values, variance_thresholds))
        # Bad variance
        variance_thresholds = {"lower": [0] * 4, "upper": [4] * 4}
        self.assertFalse(is_variance_ok(sequence.values, variance_thresholds))

    def test_whiten_multivariate(self):
        window_signal = np.array([[1, 2, 3], [1, 2, 5]], dtype=np.float64)
        expected_result = np.array([[0, 0, -1], [0, 0, 1]], dtype=np.float64)
        print(np.round(whiten_multivariate(window_signal), 2))
        self.assertTrue(np.array_equal(np.round(whiten_multivariate(window_signal), 2), expected_result))

    def test_distance_static(self):
        # TODO: write test
        pass

    def test_lb_keogh_dist(self):
        # TODO: write test
        pass

    def test_make_envelope_keogh(self):
        seq = [0, 1, 2]
        sz2 = 7
        radius = 1
        expected_res = [(1, 0), (1, 0), (1, 0), (2, 0), (2, 1), (2, 1), (2, 1)]
        dtw = DTWModel(3, 1, 7, 0.1)
        result = dtw.make_envelope_keogh(seq, sz2, radius)
        self.assertEqual(expected_res, result)

    def test_make_envelope_index_keogh(self):
        sz1 = 7
        sz2 = 3
        radius = 1
        expected_res = [(0, 2), (0, 2), (0, 2), (0, 3), (1, 3), (1, 3), (1, 4)]
        result = DTWModel.make_envelope_index_keogh(sz1, sz2, radius)
        self.assertEqual(expected_res, result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

@author: quetran
"""

import logging

from cognite_ml.timeseries.pattern_search.algorithms.DTW.searcher import DTW

logger = logging.getLogger(__name__)


class PatternSearch:
    def __init__(self, timeout=300):
        self.timeout = timeout

    def search(self, input_timeseries, pattern_timeseries, min_range=None, max_range=None, limit=10):
        searcher = DTW(min_range, max_range, limit)
        matches = searcher.search(input_timeseries, pattern_timeseries, self.timeout)

        result = {"data": {"items": matches}}
        return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

@author: quetran
"""
import collections
import logging
from functools import partial
from multiprocessing import Pool, cpu_count

import numpy as np
from cognite_ml.timeseries.pattern_search import auxiliary as aux
from cognite_ml.timeseries.pattern_search.auxiliary import find_variance_thresholds_multivariate
from cognite_ml.timeseries.pattern_search.pattern_search_base import PatternSearchBase

from . import pydtw
from .pydtw import DTWModel


logger = logging.getLogger(__name__)


class DTW(PatternSearchBase):
    def __init__(self, min_range, max_range, limit):

        PatternSearchBase.__init__(self, min_range, max_range, limit)
        # forcing to use global constraint
        self.sakoeChibaWidth = 0.1

    def _lb_to_dtw_distance(self, dict, matches, timestamp_list, query_array, input_array):
        ts_indices = aux.convert_index(
            dict["stride"], word_indices=np.arange(int(len(self.input_ts) / dict["stride"]) + 1)
        )
        start_pattern = int(timestamp_list[ts_indices[dict["i"]]])
        end_pattern = int(timestamp_list[ts_indices[dict["i"]] + dict["len"] - 1])
        if not self.overlap_th or not self._is_overlapping(matches, start_pattern, end_pattern):
            seq2 = input_array[dict["window"]]
            dtw_distance = DTWModel.distance_static(query_array, seq2, self.sakoeChibaWidth)
            return {"from": start_pattern, "to": end_pattern, "score": dtw_distance}
        return {"from": start_pattern, "to": end_pattern, "score": np.inf}

    def search(self, input_timeseries, pattern_timeseries, timeout=None):
        input_ts_cols = sorted(input_timeseries.columns.values)
        query_seqs_cols = sorted(pattern_timeseries.columns.values)

        if input_ts_cols != query_seqs_cols:
            raise ("Input data and query data must have the same column names.")

        if ("timestamp" not in input_ts_cols) or (collections.Counter(input_ts_cols)["timestamp"] != 1):
            raise ("Data must have one Timestamp column.")

        input_ts_cols.remove("timestamp")
        query_seqs_cols.remove("timestamp")

        PatternSearchBase.search_init(self, input_timeseries, pattern_timeseries, timeout)

        lower_bounds = self.find_similar_periods(
            self.query_seq.as_matrix(columns=query_seqs_cols), self.input_ts.as_matrix(columns=input_ts_cols)
        )
        sorted_lower_bounds = sorted(lower_bounds, key=lambda dict: dict["distance"])  # sort by score

        timestamp_list = self.input_ts.reset_index()["timestamp"]
        matches = []
        dtw_scores = []

        query_array = self.query_seq.as_matrix(columns=query_seqs_cols)
        input_array = self.input_ts.as_matrix(columns=input_ts_cols)

        slices = [s for s in range(0, len(sorted_lower_bounds), 3000)]
        slices.append(len(sorted_lower_bounds))

        for i in range(len(slices) - 1):
            num_cpu = cpu_count()
            with Pool(processes=num_cpu) as p:
                results = p.map_async(
                    partial(
                        self._lb_to_dtw_distance,
                        matches=matches,
                        timestamp_list=timestamp_list,
                        query_array=query_array,
                        input_array=input_array,
                    ),
                    sorted_lower_bounds[slices[i] : slices[i + 1]],
                ).get(timeout=self.timeout)
            dtw_scores.extend(results)

            dtw_scores = sorted(dtw_scores, key=lambda k: k["score"])
            min_dtw_distance = dtw_scores[0]["score"]
            if i < len(slices) - 2:
                while min_dtw_distance < sorted_lower_bounds[slices[i + 1]]["distance"]:
                    candidate_match = dtw_scores.pop(0)
                    if not self.overlap_th or not self._is_overlapping(
                        matches, candidate_match["from"], candidate_match["to"]
                    ):
                        logger.info(candidate_match)
                        matches.append(candidate_match)
                    if len(matches) == self.limit:
                        matches = aux.convert_scores_to_percentages(matches, self.max_length)
                        return matches
                    min_dtw_distance = dtw_scores[0]["score"] if dtw_scores else np.inf

        while len(matches) < self.limit:
            if len(dtw_scores) > 0:
                candidate_match = dtw_scores.pop(0)
                if not self.overlap_th or not self._is_overlapping(
                    matches, candidate_match["from"], candidate_match["to"]
                ):
                    matches.append(candidate_match)
                if len(matches) == self.limit:
                    matches = aux.convert_scores_to_percentages(matches, self.max_length)
                    return matches
            else:
                matches = aux.convert_scores_to_percentages(matches, self.max_length)
                return matches

    def _is_overlapping(self, matches, start_candidate_pattern, end_candidate_pattern):
        for match in matches:
            if (
                aux.calculate_overlap(match["from"], match["to"], start_candidate_pattern, end_candidate_pattern)
                > self.overlap_th
            ):
                return True
        return False

    def find_similar_patterns(self, search_len, expected_seq, timeseries):
        stride = int(search_len / 20) + 1
        dtw_model = DTWModel(search_len, stride, len(expected_seq), self.sakoeChibaWidth)
        windows = list(dtw_model.sliding_window_index(len(timeseries)))
        variance_thresholds = find_variance_thresholds_multivariate(expected_seq)
        distances = []
        new_expected_seq = pydtw.whiten_multivariate(expected_seq).transpose()
        envelope = []
        l = len(new_expected_seq)
        for i in range(l):
            envelope.append(
                dtw_model.make_envelope_keogh(
                    new_expected_seq[i], search_len, max(1, self.sakoeChibaWidth * max(len(expected_seq), search_len))
                )
            )
        envelope = np.array(envelope)
        for i in range(len(windows)):
            if not self.filterVariance or pydtw.is_variance_ok(timeseries[windows[i]], variance_thresholds):
                distances.append(
                    {
                        "distance": dtw_model.lb_keogh_dist(expected_seq, timeseries[windows[i]], envelope),
                        "stride": stride,
                        "len": search_len,
                        "i": i,
                        "window": windows[i],
                    }
                )
        return distances
